// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "SimpleTexturedMaterial.hpp"
#include "util/JSON.hpp"
#include "asset/loader/TextureLoader.hpp"
#include "asset/loader/ShaderLoader.hpp"

using namespace Dawn;

void SimpleTexturedMaterial::initShaderPrograms() {
  if(this->shader) return;

  auto shaderFile = getGame()->assetManager->get<ShaderLoader>("shaders/simple-textured.slang");
  this->shader = shaderFile->getShader();

  auto structure = this->shader->getStructure();
  this->data = structure->createData();

  this->setColor(COLOR_WHITE);
  this->setTexture(nullptr);
}

void SimpleTexturedMaterial::load(std::shared_ptr<SceneLoadContext> ctx) {
  this->initShaderPrograms();

  if(ctx->data.contains("color")) {
    this->setColor(JSON::color(ctx->data["color"]));
  }

  if(ctx->data.contains("texture")) {
    auto asset = ctx->getAsset<TextureLoader>(
      ctx->data["texture"].get<std::string>()
    );
    this->setTexture(asset->getTexture());
  }
}

struct Color SimpleTexturedMaterial::getColor() {
  auto structure = this->shader->getStructure();
  return data->get<struct Color>(structure->getOffset("color"));
}

std::shared_ptr<Texture> SimpleTexturedMaterial::getTexture() {
  auto structure = this->shader->getStructure();
  return data->get<std::shared_ptr<Texture>>(structure->getOffset("texture"));
}

void SimpleTexturedMaterial::setTexture(
  const std::shared_ptr<Texture> texture
) {
  auto structure = this->shader->getStructure();
  data->set<std::shared_ptr<Texture>>(structure->getOffset("texture"), texture);
  data->set<bool_t>(structure->getOffset("hasTexture"), texture != nullptr);
}

void SimpleTexturedMaterial::setColor(const struct Color color) {
  auto structure = this->shader->getStructure();
  data->set<struct Color>(structure->getOffset("color"), color);
}

std::vector<std::shared_ptr<RenderPass>> SimpleTexturedMaterial::getPasses(
  struct RenderPassContext &ctx
) {
  auto structure = this->shader->getStructure();

  data->set<glm::mat4>(
    structure->getOffset("projection"),
    ctx.camera->getProjection()
  );

  data->set<glm::mat4>(
    structure->getOffset("view"),
    ctx.camera->getItem()->getWorldTransform()
  );

  data->set<glm::mat4>(
    structure->getOffset("model"),
    this->getItem()->getWorldTransform()
  );

  return {
    std::make_shared<RenderPass>(
      *this,
      nullptr,// Get mesh automatically.
      MeshDrawMode::TRIANGLES,// Move this later.
      0,// Move this later.
      -1,// Move this later.
      this->shader,
      this->data
    )
  };
}
