// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "IShaderData.hpp"

using namespace Dawn;

IShaderData::IShaderData(
  const std::shared_ptr<ShaderStructure> &structure
) : structure(structure) {
  data = (void *)malloc(structure->size);
}

template<>
void *IShaderData::get<void *>(const size_t offset) {
  return (void *)((size_t)data + offset);
}

template<>
int8_t IShaderData::get<int8_t>(const size_t offset) {
  return *(int8_t*)this->get<void*>(offset);
}

template<>
int16_t IShaderData::get<int16_t>(const size_t offset) {
  return *(int16_t *)this->get<void *>(offset);
}

template<>
int32_t IShaderData::get<int32_t>(const size_t offset) {
  return *(int32_t *)this->get<void *>(offset);
}

template<>
int64_t IShaderData::get<int64_t>(const size_t offset) {
  return *(int64_t *)this->get<void *>(offset);
}

template<>
uint8_t IShaderData::get<uint8_t>(const size_t offset) {
  return *(uint8_t *)this->get<void *>(offset);
}

template<>
uint16_t IShaderData::get<uint16_t>(const size_t offset) {
  return *(uint16_t *)this->get<void *>(offset);
}

template<>
uint32_t IShaderData::get<uint32_t>(const size_t offset) {
  return *(uint32_t *)this->get<void *>(offset);
}

template<>
uint64_t IShaderData::get<uint64_t>(const size_t offset) {
  return *(uint64_t *)this->get<void *>(offset);
}

template<>
float_t IShaderData::get<float_t>(const size_t offset) {
  return *(float_t *)this->get<void *>(offset);
}

template<>
glm::vec2 IShaderData::get<glm::vec2>(const size_t offset) {
  return *(glm::vec2 *)this->get<void *>(offset);
}

template<>
glm::vec3 IShaderData::get<glm::vec3>(const size_t offset) {
  return *(glm::vec3 *)this->get<void *>(offset);
}

template<>
glm::vec4 IShaderData::get<glm::vec4>(const size_t offset) {
  return *(glm::vec4 *)this->get<void *>(offset);
}

template<>
glm::mat2 IShaderData::get<glm::mat2>(const size_t offset) {
  return *(glm::mat2 *)this->get<void *>(offset);
}

template<>
glm::mat3 IShaderData::get<glm::mat3>(const size_t offset) {
  return *(glm::mat3 *)this->get<void *>(offset);
}

template<>
glm::mat4 IShaderData::get<glm::mat4>(const size_t offset) {
  return *(glm::mat4 *)this->get<void *>(offset);
}

template<>
bool_t IShaderData::get<bool_t>(const size_t offset) {
  return this->get<int32_t>(offset) != 0;
}

template<>
struct Color IShaderData::get<struct Color>(const size_t offset) {
  glm::vec4 v = this->get<glm::vec4>(offset);
  return {
    v.r,
    v.g,
    v.b,
    v.a
  };
}



void IShaderData::setRaw(
  const size_t offset,
  const void *value,
  const size_t size
) {
  memcpy(this->get<void *>(offset), value, size);
  this->dirty = true;
}

template<>
void IShaderData::set<int8_t>(const size_t offset, const int8_t value) {
  this->setRaw(offset, &value, sizeof(int8_t));
}

template<>
void IShaderData::set<int16_t>(const size_t offset, const int16_t value) {
  this->setRaw(offset, &value, sizeof(int16_t));
}

template<>
void IShaderData::set<int32_t>(const size_t offset, const int32_t value) {
  this->setRaw(offset, &value, sizeof(int32_t));
}

template<>
void IShaderData::set<int64_t>(const size_t offset, const int64_t value) {
  this->setRaw(offset, &value, sizeof(int64_t));
}

template<>
void IShaderData::set<uint8_t>(const size_t offset, const uint8_t value) {
  this->setRaw(offset, &value, sizeof(uint8_t));
}

template<>
void IShaderData::set<uint16_t>(const size_t offset, const uint16_t value) {
  this->setRaw(offset, &value, sizeof(uint16_t));
}

template<>
void IShaderData::set<uint32_t>(const size_t offset, const uint32_t value) {
  this->setRaw(offset, &value, sizeof(uint32_t));
}

template<>
void IShaderData::set<uint64_t>(const size_t offset, const uint64_t value) {
  this->setRaw(offset, &value, sizeof(uint64_t));
}

template<>
void IShaderData::set<float_t>(const size_t offset, const float_t value) {
  this->setRaw(offset, &value, sizeof(float_t));
}

template<>
void IShaderData::set<glm::vec2>(const size_t offset, const glm::vec2 value) {
  this->setRaw(offset, &value, sizeof(glm::vec2));
}

template<>
void IShaderData::set<glm::vec3>(const size_t offset, const glm::vec3 value) {
  this->setRaw(offset, &value, sizeof(glm::vec3));
}

template<>
void IShaderData::set<glm::vec4>(const size_t offset, const glm::vec4 value) {
  this->setRaw(offset, &value, sizeof(glm::vec4));
}

template<>
void IShaderData::set<glm::mat2>(const size_t offset, const glm::mat2 value) {
  this->setRaw(offset, &value, sizeof(glm::mat2));
}

template<>
void IShaderData::set<glm::mat3>(const size_t offset, const glm::mat3 value) {
  this->setRaw(offset, &value, sizeof(glm::mat3));
}

template<>
void IShaderData::set<glm::mat4>(const size_t offset, const glm::mat4 value) {
  this->setRaw(offset, &value, sizeof(glm::mat4));
}

template<>
void IShaderData::set<struct Color>(
  const size_t offset,
  const struct Color value
) {
  this->set<glm::vec4>(offset, glm::vec4(
    value.r,
    value.g,
    value.b,
    value.a
  ));
}

template<>
void IShaderData::set<bool_t>(const size_t offset, const bool_t value) {
  this->set<>(offset, value ? 1 : 0);
}

IShaderData::~IShaderData() {
  if(data != nullptr) {
    free(data);
    data = nullptr;
  }
}