// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "IShaderProgram.hpp"
#include "assert/assert.hpp"
#if DAWN_DEBUG_SHADERS
  #include <fstream>
#endif

using namespace Dawn;

void IShaderProgram::init(
  IModule *module,
  Slang::ComPtr<ISession> session
) {
  assertNotNull(module, "Module cannot be null.");
  this->module = module;

  // Get list of entry points and create components
  int32_t definedEntryPointCount = module->getDefinedEntryPointCount();
  // +1 for module
  auto components = new IComponentType*[definedEntryPointCount + 1];
  int32_t componentCount = 0;
  components[componentCount++] = module;// First component is module.

  // Get the entry point info and append to components list.
  for(auto i = 0; i < definedEntryPointCount; i++) {
    auto result = module->getDefinedEntryPoint(
      i,
      (IEntryPoint**)(&components[componentCount++])
    );
    if(result != SLANG_OK) assertUnreachable("Failed to get entry point.");
  }

  // Create the composite component type
  Slang::ComPtr<IBlob> diagnostics;
  session->createCompositeComponentType(
    components,
    componentCount,
    program.writeRef(),
    diagnostics.writeRef()
  );
  if(diagnostics) {
    assertUnreachable("%s\n", (const char*) diagnostics->getBufferPointer());
    return;
  }

  // Link the program.
  auto result = program->link(linkedProgram.writeRef(), diagnostics.writeRef());
  if(diagnostics) {
    assertUnreachable("%s\n", (const char*) diagnostics->getBufferPointer());
    return;
  }

  // Get the layout
  layout = program->getLayout();

  // Create the shader program.
  Slang::ComPtr<IBlob> blob;
  for(auto i = 0; i < definedEntryPointCount; i++) {
    // Get the code
    auto result = linkedProgram->getEntryPointCode(
      i,
      0,
      blob.writeRef(),
      diagnostics.writeRef()
    );

    if(diagnostics) {
      assertUnreachable("%s\n", (const char*) diagnostics->getBufferPointer());
    }

    // Get the stage information
    auto entryPointReflection = layout->getEntryPointByIndex(i);
    auto stage = entryPointReflection->getStage();

    // Write out to file for debugging
    #if DAWN_DEBUG_SHADERS
      std::filesystem::path filePath = (
        "debug/" +
        std::string(module->getName()) +
        "/" +
        std::string(entryPointReflection->getName()) +
        ".glsl"
      );
      std::filesystem::create_directories(filePath.parent_path());
      std::cout << "Writing shader to " << filePath << std::endl;
      std::ofstream file(filePath);
      file << (const char*)blob->getBufferPointer();
      file.close();
    #endif

    // Create the shader entry
    auto shaderStage = std::make_shared<ShaderStage>();
    shaderStage->init(
      stage,
      std::string((const char*)blob->getBufferPointer())
    );

    // Add to the list
    stages.push_back(shaderStage);
  }

  // Cleanup components
  delete [] components;

  // Reflect out the data
  auto rootLayout = layout->getGlobalParamsVarLayout();
  auto rootLayoutTypeLayout = rootLayout->getTypeLayout();
  assertTrue(
    rootLayoutTypeLayout->getKind() == TypeReflection::Kind::ConstantBuffer,
    "Root layout, should and must be a constant buffer."
  );

  auto realRootLayout = rootLayoutTypeLayout->getElementVarLayout();
  auto realRootLayoutTypeLayout = realRootLayout->getTypeLayout();
  assertTrue(
    realRootLayoutTypeLayout->getKind() == TypeReflection::Kind::Struct,
    "Real root layout should and must be a struct."
  );

  // Now we can really begin parsing out the data.
  structure = std::make_shared<struct ShaderStructure>(
    realRootLayout->getTypeLayout(),
    "root",
    0
  );
}

std::shared_ptr<struct ShaderStructure> IShaderProgram::getStructure() {
  return structure;
}

IShaderProgram::~IShaderProgram() {
  // Release the linked program
  if(linkedProgram) {
    linkedProgram->release();
    linkedProgram = nullptr;
  }
}