// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "display/shader/ShaderStage.hpp"
#include "display/shader/ShaderData.hpp"

namespace Dawn {

  class IShaderProgram {
    protected:
      IModule *module;
      std::vector<std::shared_ptr<ShaderStage>> stages;
      ComPtr<IComponentType> program;
      ComPtr<IComponentType> linkedProgram;
      ProgramLayout *layout;
      std::shared_ptr<struct ShaderStructure> structure;

    public:
      /**
       * Initialize the ShaderProgram. In your render hosts' implementation
       * this would initialize the shader on your GPU and ideally set the
       * initial values for uniforms and objects.
       * 
       * Provided stages will already be "initialized" (at least as far as the
       * engine is concerned), and should be ready to be used. If your render
       * host needs the program to be initialized prior to the stages you will
       * need to modify how stages are initialized, this is by design as we are
       * loading the shader code and "initializing" the stages before we create
       * and initialize the program which will use and link the stages.
       * 
       * @param module The SLANG module that was loaded.
       * @param session The SLANG session that was used to load the module.
       */
      virtual void init(
        IModule *module,
        ComPtr<ISession> session
      );

      /**
       * Returns the structure of the shader program.
       * 
       * @return The structure of the shader program.
       */
      std::shared_ptr<struct ShaderStructure> getStructure();

      /**
       * Binds this shader program as the currently active shader program.
       */
      virtual void bind() = 0;

      /**
       * Destroy the IShaderProgram2 object
       */
      virtual ~IShaderProgram();
  };
}