// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "ShaderStructure.hpp"
#include "display/shader/ShaderData.hpp"
#include "assert/assert.hpp"

using namespace Dawn;

ShaderStructure::ShaderStructure(
  TypeLayoutReflection *typeLayout,
  const std::string name,
  const size_t start
) {
  this->name = name;
  this->start = start;
  this->size = typeLayout->getSize();
  this->alignment = typeLayout->getAlignment();

  switch(typeLayout->getKind()) {
    case TypeReflection::Kind::Struct: {
      this->type = ShaderStructureType::STRUCT;
      auto count = typeLayout->getFieldCount();
      size_t offset = this->start;
      for(auto i = 0; i < count; i++) {
        auto field = typeLayout->getFieldByIndex(i);
        auto cName = field->getName();
        auto member = std::make_shared<ShaderStructure>(
          field->getTypeLayout(),
          cName ? std::string(cName) : "",
          offset
        );
        offset += member->size;
        this->members.push_back(member);
      }
      break;
    }

    case TypeReflection::Kind::Array: {
      this->type = ShaderStructureType::ARRAY;
      this->length = typeLayout->getElementCount();
      auto elementType = typeLayout->getElementTypeLayout();
      for(auto i = 0; i < this->length; i++) {
        auto member = std::make_shared<ShaderStructure>(
          elementType,
          std::to_string(i),
          this->start + (i * elementType->getSize())
        );
        this->members.push_back(member);
      }
      break;
    }

    case TypeReflection::Kind::Matrix:
    case TypeReflection::Kind::Vector:
    case TypeReflection::Kind::Scalar:
      this->type = ShaderStructureType::VARIABLE;
      break;

    case TypeReflection::Kind::Resource:
      switch(typeLayout->getResourceShape()) {
        case SlangResourceShape::SLANG_TEXTURE_2D:
          this->type = ShaderStructureType::TEXTURE2D;
          break;

        default:
          assertUnreachable(
            "Unknown resource shape: %d",
            typeLayout->getResourceShape()
          );
          break;
      }
      break;

    default: {
      assertUnreachable("Unknown type layout kind: %d", typeLayout->getKind());
    }
  }
}

std::shared_ptr<ShaderStructure> ShaderStructure::getStructMember(
  const std::string &name
) {
  assertTrue(
    this->type == ShaderStructureType::STRUCT,
    "This structure is not a struct."
  );
  for(auto &member : this->members) {
    if(member->name == name) return member;
  }
  assertUnreachable("Struct member not found: %s", name.c_str());
  return nullptr;
}

std::shared_ptr<ShaderStructure> ShaderStructure::getArrayMember(
  size_t index
) {
  assertTrue(
    this->type == ShaderStructureType::ARRAY,
    "This structure is not an array."
  );
  return this->members.at(index);
}

size_t ShaderStructure::getOffset() {
  return this->start;
}

size_t ShaderStructure::getOffset(const std::string &name) {
  return this->getStructMember(name)->getOffset();
}

size_t ShaderStructure::getOffset(size_t index) {
  return this->getArrayMember(index)->getOffset();
}

std::shared_ptr<ShaderData> ShaderStructure::createData() {
  return std::make_shared<ShaderData>(shared_from_this());
}