// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "display/shader/ShaderData.hpp"
#include "assert/assert.hpp"
#include "assert/assertgl.hpp"

using namespace Dawn;

ShaderData::ShaderData(
  const std::shared_ptr<ShaderStructure> &structure
) : IShaderData(structure) {
  glGenBuffers(1, &this->buffer);
  assertNoGLError();

  // Quickly scan the structure for textures that we can bind to.
  int32_t nextTexture = 0;
  std::function<void(const std::shared_ptr<ShaderStructure>&)>
    recursivelyAppendTextures
  ;
  recursivelyAppendTextures = [&](
    const std::shared_ptr<ShaderStructure> &structure
  ) {
    assertTrue(
      structure->type == ShaderStructureType::STRUCT,
      "Can only scan for texture indices in structures."
    );

    for(auto &field : structure->members) {
      switch(field->type) {
        case ShaderStructureType::STRUCT:
          recursivelyAppendTextures(field);
          break;
        
        case ShaderStructureType::TEXTURE2D:
          this->textureIndices[field->getOffset()] = nextTexture++;
          break;

        default:
          break;
      }
    }
  };
  recursivelyAppendTextures(structure);

}

void ShaderData::upload() {
  if(!this->dirty) return;

  glBindBuffer(GL_UNIFORM_BUFFER, this->buffer);
  assertNoGLError();
  this->bind();
  glBufferData(GL_UNIFORM_BUFFER, this->structure->size, this->data, GL_STATIC_DRAW);
  assertNoGLError();

  for(auto &pair : this->textureBindings) {
    if(!pair.second) continue;
    pair.second->bind(pair.first);
    glUniform1i(pair.first, pair.first);
  }
  
  this->dirty = false;
}

void ShaderData::bind() {
  glBindBufferBase(GL_UNIFORM_BUFFER, 0, this->buffer);
  assertNoGLError();
}

template<>
std::shared_ptr<Texture> IShaderData::get<std::shared_ptr<Texture>>(
  const size_t offset
) {
  ShaderData *self = (ShaderData *)this;
  auto map = self->textureIndices.find(offset);
  assertTrue(
    map != self->textureIndices.end(),
    "Texture is not found in the structure or the offset is incorrect."
  );

  auto index = map->second;
  return self->textureBindings[index];
}

template<>
void IShaderData::set<std::shared_ptr<Texture>>(
  const size_t offset, const std::shared_ptr<Texture> value
) {
  ShaderData *self = (ShaderData *)this;
  auto map = self->textureIndices.find(offset);
  assertTrue(
    map != self->textureIndices.end(),
    "Texture is not found in the structure or the offset is incorrect."
  );

  // I didn't implement it, but in theory we can check if the texture is
  // already bound to an index and reuse its existing index to avoid
  // binding a texture twice if we don't need to, however this would
  // require all bindings be flushed and re-bound each frame.
  auto index = map->second;
  self->textureBindings[map->second] = value;// Releases textures too.
  if(value) this->dirty = true;
}

ShaderData::~ShaderData() {
  if(this->buffer != -1) {
    glDeleteBuffers(1, &this->buffer);
    assertNoGLError();
    this->buffer = -1;
  }
}